﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Net;
using System.Net.Sockets;

namespace rdeclient
{
    public partial class rdeclient : Form
    {
        const int kBufferSize = 4096;
        const string kListFile = "list.txt";

        public delegate void DelegateAddConsoleMessage(string msg);

        public rdeclient()
        {
            InitializeComponent();
        }

        private void button1_Click(object sender, EventArgs e)
        {
            try
            {
                if (m_socket != null && m_socket.Connected)
                {
                    SendSocketClosedMessage();
                    CloseSocket();
                    return;
                }

                ConnectDialog dlg = new ConnectDialog(rde.Common.kDefaultPort);
                if (dlg.ShowDialog(this) != DialogResult.OK)
                    return;

                m_socket = new Socket(AddressFamily.InterNetwork, SocketType.Stream,
                    ProtocolType.Tcp);
                IPAddress ip = IPAddress.Parse(dlg.GetServer());
                int port = dlg.GetPort();
                IPEndPoint ipEnd = new IPEndPoint(ip, port);
                m_socket.Connect(ipEnd);
                if (m_socket.Connected)
                {
                    SendMachineName();
                    button1.Text = "Disconnect Worker";
                    AddConsoleMessage("Connected to " + dlg.GetServer() + " at port " + port.ToString());
                    WaitForData();
                }
            }
            catch (SocketException exc)
            {
                MessageBox.Show(exc.Message);
                m_socket = null;
            }
        }

        class SocketPacket
        {
            public byte[] dataBuffer;
        };

        void WaitForData()
        {
            try
            {
                if (m_dataCallback == null)
                    m_dataCallback = new AsyncCallback(OnDataReceived);

                SocketPacket packet = new SocketPacket();
                packet.dataBuffer = new byte[kBufferSize];
                m_socket.BeginReceive(packet.dataBuffer, 0, packet.dataBuffer.Length,
                    SocketFlags.None, m_dataCallback, packet);
            }
            catch (SocketException exc)
            {
                MessageBox.Show(exc.Message);
            }
        }

        // String with length (2 bytes of len, then string).
        string CustomStringFromData(byte[] msg, ref int index)
        {
            int strLen = (msg[index] << 8) | msg[index + 1];
            System.Diagnostics.Debug.Assert(strLen >= 0 && strLen < 8192);
            string ret = null;

            if (strLen > 0)
            {
                char[] chars = new char[strLen];
                System.Text.Decoder d = System.Text.Encoding.ASCII.GetDecoder();
                d.GetChars(msg, index + 2, strLen, chars, 0);
                ret = new string(chars);
                index += strLen;
            }
            index += 2;
            return ret;
        }

        public class GroupDesc
        {
            public string outDirectory;
            public string commandLine;
        };

        void ProcessForGroup(int index)
        {
            TextWriter tw = new StreamWriter(kListFile, false);
            int numFilesInGroup = 0;
            string firstFile = null;
            foreach (FileDesc desc in m_filesToProcess)
            {
                if (desc.groupIndex == index)
                {
                    if (numFilesInGroup == 0)
                        firstFile = desc.fileName;
                    tw.WriteLine(desc.fileName);
                    ++numFilesInGroup;
                }
            }
            tw.Close();
            Invoke(new DelegateAddConsoleMessage(AddConsoleMessage),
                    new Object[] { "Processing " + numFilesInGroup.ToString() + " file(s) in group " + index.ToString() });
            if (numFilesInGroup > 0)
            {
                string cmdLine = ReplaceMacros(m_groups[index].commandLine, firstFile, index);
                ExecuteTool(cmdLine);
            }
        }

        void HandleMessage(byte[] msg)
        {
            rde.Common.ServerToClientMsg msgType = (rde.Common.ServerToClientMsg)msg[0];
            if (msgType == rde.Common.ServerToClientMsg.FILE_NAME)
            {
                // For more responsive UI we'd probably should process files
                // on another thread, but this will do for now.
                if (m_state == rde.Common.ClientState.WORKING)
                {
                    byte groupIndex = msg[1];
                    string fileName = rde.Common.GetASCIIZStringFromData(msg, 2);
                    if (m_workingMode == rde.Common.ClientWorkingMode.BATCH)
                    {
                        m_filesToProcess.Add(new FileDesc(fileName, groupIndex));
                        // Give other clients chance to grab something.
                        // Very "straightforward" way.
                        // It simulates launching the tool for batch mode.
                        // Surely can be done in more intelligent manner.
                        System.Threading.Thread.Sleep(200 + m_rand.Next(200));
                    }
                    else
                    {
                        Invoke(new DelegateAddConsoleMessage(AddConsoleMessage),
                            new Object[] { "Processing " + fileName + " " + groupIndex.ToString() });

                        string arg = ReplaceMacros(m_groups[groupIndex].commandLine, fileName, groupIndex);
                        ExecuteTool(arg);
                    }
                    SendSimpleMessage(rde.Common.ClientToServerMsg.ASK_FOR_WORK);
                }
            }
            else if (msgType == rde.Common.ServerToClientMsg.TOOL_EXECUTABLE)
            {
                m_ignoreErrors = msg[1] != 0;
                m_tool = rde.Common.GetASCIIZStringFromData(msg, 2);
                Invoke(new DelegateAddConsoleMessage(AddConsoleMessage),
                    new Object[] { "Tool executable: " + m_tool });
            }
            else if (msgType == rde.Common.ServerToClientMsg.START_PROCESS)
            {
                int i = 1;
                string exec = CustomStringFromData(msg, ref i);
                string args = CustomStringFromData(msg, ref i);
                Process prc = Process.Start(exec, args);
                if (prc != null)
                    prc.WaitForExit();
            }
            else if (msgType == rde.Common.ServerToClientMsg.WORKING_MODE)
            {
                string[] workingModes = { "Load balancing", "Batch" };
                m_workingMode = (rde.Common.ClientWorkingMode)msg[1];
                m_state = rde.Common.ClientState.WORKING;
                SendSocketStateMessage();
                SendSimpleMessage(rde.Common.ClientToServerMsg.ASK_FOR_WORK);
                Invoke(new DelegateAddConsoleMessage(AddConsoleMessage),
                    new Object[] { "Working mode set to " + workingModes[msg[1]] });
            }
            else if (msgType == rde.Common.ServerToClientMsg.FILE_LIST_COMPLETE)
            {
                if (m_workingMode == rde.Common.ClientWorkingMode.BATCH)
                {
                    for (int i = 0; i < m_groups.Count; ++i)
                        ProcessForGroup(i);
                }
                System.Diagnostics.Debug.Assert(m_filesToProcess.Count == 0 ||
                    m_workingMode == rde.Common.ClientWorkingMode.BATCH);
                m_filesToProcess.Clear();
                m_state = rde.Common.ClientState.IDLE;
                SendSocketStateMessage();

                if (m_postBuildStep != null)
                {
                    Process prc = Process.Start(m_postBuildStep, m_postBuildArgs);
                    if (prc != null)
                        prc.WaitForExit();
                }

                Invoke(new DelegateAddConsoleMessage(AddConsoleMessage),
                    new Object[] { "Finished!" });
                m_groups.Clear();

                SendSimpleMessage(rde.Common.ClientToServerMsg.TASK_DONE);
            }
            else if (msgType == rde.Common.ServerToClientMsg.GROUP_DESCRIPTION)
            {
                GroupDesc desc = new GroupDesc();
                int i = 1;
                desc.outDirectory = CustomStringFromData(msg, ref i);
                desc.commandLine = CustomStringFromData(msg, ref i);
                m_groups.Add(desc);
            }
            else if (msgType == rde.Common.ServerToClientMsg.TASK_DESCRIPTION)
            {
                int i = 1;
                m_ignoreErrors = msg[i++] != 0;
                m_destExtension = CustomStringFromData(msg, ref i);
                m_workingDirectory = CustomStringFromData(msg, ref i);
                m_tool = CustomStringFromData(msg, ref i);
                m_postBuildStep = CustomStringFromData(msg, ref i);
                m_postBuildArgs = CustomStringFromData(msg, ref i);
                m_groups.Clear();
                Invoke(new DelegateAddConsoleMessage(AddConsoleMessage),
                    new Object[] { "Tool executable: " + m_tool });
            }
            else if (msgType == rde.Common.ServerToClientMsg.BREAK)
            {
                if (m_toolProcess != null)
                    m_toolProcess.Kill();
                m_toolProcess = null;
                m_filesToProcess.Clear();
                m_state = rde.Common.ClientState.IDLE;
                SendSocketStateMessage();
                Invoke(new DelegateAddConsoleMessage(AddConsoleMessage),
                    new Object[] { "Interrupted!" });
            }
            else
            {
                System.Diagnostics.Debug.Fail("Unknown command ID: " + msgType);
            }
        }

        void OnDataReceived(IAsyncResult res)
        {
            try
            {
                SocketPacket packet = (SocketPacket)res.AsyncState;
                int receivedBytes = m_socket.EndReceive(res);

                List<byte[]> messages =
                    m_msgReceiver.OnDataReceived(packet.dataBuffer, receivedBytes);
                for (int i = 0; i < messages.Count; ++i)
                {
                    HandleMessage(messages[i]);
                }
                if (m_socket != null && m_socket.Connected)
                    WaitForData();
            }
            catch (NullReferenceException)
            {
                // Socket closed most probably...
            }
            catch (ObjectDisposedException)
            {
                // Socket closed...
            }
            catch (SocketException exc)
            {
                MessageBox.Show(exc.Message);
                CloseSocket();
            }
        }

        void CloseSocket()
        {
            if (m_socket != null)
            {
                m_socket.Shutdown(SocketShutdown.Both);
                m_socket.Close();
                m_socket = null;
                Invoke(new DelegateAddConsoleMessage(AddConsoleMessage),
                    new Object[] { "Disconnected" });
            }
        }

        void ExecuteTool(string arg)
        {
            Invoke(new DelegateAddConsoleMessage(AddConsoleMessage), new Object[] { "Cmdline: " + arg });
            ProcessStartInfo psi = new ProcessStartInfo(m_tool, arg);
            //if (!m_ignoreErrors)
            {
                psi.RedirectStandardError = true;
                //psi.RedirectStandardOutput = true;
                psi.UseShellExecute = false;
            }
            Process prc = Process.Start(psi);
            if (prc == null)
            {
                SendToolErrorMessage("Cannot execute " + m_tool);
                return;
            }
            string errMsg = prc.StandardError.ReadToEnd();
            //string outP = prc.StandardOutput.ReadToEnd();
            m_toolProcess = prc;
            prc.WaitForExit();
            //Invoke(new DelegateAddConsoleMessage(AddConsoleMessage),
            //  new Object[] { outP });

            if (prc.ExitCode != 0)// && !m_ignoreErrors)
            {
                //System.Diagnostics.Debug.Fail("DOH, exit code fucked up");
                errMsg = "Error [" + arg + "]: " + errMsg;
                SendToolErrorMessage(errMsg);
                Invoke(new DelegateAddConsoleMessage(AddConsoleMessage),
                    new Object[] { errMsg });
            }
            m_toolProcess = null;
        }

        void SendMachineName()
        {
            string clientName = System.Environment.MachineName;
            byte[] data = new byte[clientName.Length + 2];
            data[0] = (byte)rde.Common.ClientToServerMsg.MACHINE_NAME;
            System.Text.Encoding.ASCII.GetBytes(clientName, 0, clientName.Length,
                data, 1);

            if (!rde.Common.SendMessage(m_socket, data))
                m_socket = null;
        }
        void SendSimpleMessage(rde.Common.ClientToServerMsg msgId)
        {
            byte[] data = new byte[1];
            data[0] = (byte)msgId;
            rde.Common.SendMessage(m_socket, data);
        }
        void SendSocketStateMessage()
        {
            byte[] data = new byte[2];
            data[0] = (byte)rde.Common.ClientToServerMsg.CLIENT_STATE;
            data[1] = (byte)m_state;
            rde.Common.SendMessage(m_socket, data);
        }
        void SendSocketClosedMessage()
        {
            byte[] data = new byte[1];
            data[0] = (byte)rde.Common.ClientToServerMsg.CLIENT_CLOSING;
            if (m_socket != null && m_socket.Connected)
            {
                rde.Common.SendMessage(m_socket, data);
            }
        }
        void SendToolErrorMessage(string errMsg)
        {
            byte[] data = new byte[errMsg.Length + 2];
            data[0] = (byte)rde.Common.ClientToServerMsg.TOOL_ERROR;
            System.Text.Encoding.ASCII.GetBytes(errMsg, 0, errMsg.Length, data, 1);
            rde.Common.SendMessage(m_socket, data);
        }

        public class FileDesc
        {
            public FileDesc(string name, int group)
            {
                fileName = name;
                groupIndex = group;
            }
            public string fileName;
            public int groupIndex;
        };

        Socket m_socket = null;
        AsyncCallback m_dataCallback = null;
        string m_tool;
        List<FileDesc> m_filesToProcess = new List<FileDesc>();
        rde.MsgReceiver m_msgReceiver = new rde.MsgReceiver();
        bool m_connected = false;
        bool m_ignoreErrors = false;
        List<GroupDesc> m_groups = new List<GroupDesc>();
        rde.Common.ClientWorkingMode m_workingMode = rde.Common.ClientWorkingMode.BATCH;
        rde.Common.ClientState m_state = rde.Common.ClientState.IDLE;
        string m_workingDirectory = "";
        string m_destExtension = "";
        Process m_toolProcess = null;
        string m_postBuildStep = null;
        string m_postBuildArgs = null;
        Random m_rand = new Random();

        // groupIndex == -1 for "global" replace (ie, not dependent on the entry group).
        string ReplaceMacros(string commandLine, string fileName, int groupIndex)
        {
            if (commandLine == null)
                return fileName;

            // $$$ TODO: possible optimization, only create file info if it's
            // $$$ needed.
            FileInfo fi = new FileInfo(fileName);
            commandLine = commandLine.Replace("$(SourcePath)", fileName);
            commandLine = commandLine.Replace("$(SourceFile)", fi.Name);
            commandLine = commandLine.Replace("$(SourceExt)", fi.Extension);
            commandLine = commandLine.Replace("$(SourceDir)", fi.DirectoryName);
            commandLine = commandLine.Replace("$(WorkingDir)", m_workingDirectory);

            string outDir = groupIndex >= 0 ? m_groups[groupIndex].outDirectory : null;
            if (outDir == null)
            {
                outDir = fi.DirectoryName;
            }
            else
                outDir = m_workingDirectory + "\\" + outDir;

            string outPath = outDir + "\\" + fi.Name;
            outPath = outPath.Replace(fi.Extension, m_destExtension);
            commandLine = commandLine.Replace("$(OutPath)", outPath);
            string outFile = fi.Name.Replace(fi.Extension, m_destExtension);
            commandLine = commandLine.Replace("$(OutFile)", outFile);
            commandLine = commandLine.Replace("$(OutDir)", outDir);
            commandLine = commandLine.Replace("$(OutExt)", m_destExtension);
            commandLine = commandLine.Replace("$(ListFile)", kListFile);

            return commandLine;
        }

        private void rdeclient_Load(object sender, EventArgs e)
        {

        }

        private void rdeclient_FormClosing(object sender, FormClosingEventArgs e)
        {
            SendSocketClosedMessage();
            CloseSocket();
        }

        void AddConsoleMessage(string msg)
        {
            textBoxConsole.Text += msg + "\r\n";
            // This seems to be common way to scroll text to the bottom.
            // http://bytes.com/forum/thread255749.html
            // Not very elegant, but works.
            textBoxConsole.SelectionStart = textBoxConsole.Text.Length;
            textBoxConsole.SelectionLength = 0;
            textBoxConsole.ScrollToCaret();
        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            if (m_connected)
            {
                if (m_socket == null || !m_socket.Connected)
                {
                    button1.Text = "Connect As Worker";
                    m_connected = false;
                }
            }
            else
            {
                if (m_socket != null && m_socket.Connected)
                    m_connected = true;
            }
        }
    }
}
