﻿using System;
using System.Collections.Generic;
using System.IO;    
using System.Linq;
using System.Text;
using System.Xml.Serialization;
using System.Net.Sockets;

namespace rdeserver
{
    [Serializable]
    public class Task
    {
        public void ExecutePreBuild()
        {
            if (preBuildStep.Length > 0)
            {
                System.Diagnostics.ProcessStartInfo psi =
                    new System.Diagnostics.ProcessStartInfo(preBuildStep);
                psi.RedirectStandardOutput = true;
                System.Diagnostics.Process prc = System.Diagnostics.Process.Start(psi);
                prc.WaitForExit();
            }
        }
        public void PrepareForExecution()
        {
            List<EntryGroup> originalGroups = new List<EntryGroup>(entryGroups.ToArray());
            foreach (EntryGroup group in originalGroups)
            {
                // groups with non-empty "dir" are "fake" directories, created only
                // to batch output files.
                if (group.dir.Length == 0)  
                    BuildSourceFilesList(group);
            }
        }
        public void AddGroup(EntryGroup group)
        {
            group.id = entryGroups.Count;
            entryGroups.Add(group);
        }

        static bool ContainsEntryWithName(List<Entry> lst, string fname)
        {
            foreach (Entry e in lst)
            {
                if (e.fileName == fname)
                    return true;
            }
            return false;
        }

        int FindGroupIdContainingDir(string dir)
        {
            for (int i = 0; i < entryGroups.Count; ++i)
            {
                if (entryGroups[i].dir == dir)
                    return i;
            }
            return -1;
        }
        void BuildSourceFilesList(EntryGroup group)
        {
            if (group.files != null)
            {
                foreach (string fileName in group.files)
                {
                    FileInfo fileInfo = new FileInfo(fileName);
                    if (fileInfo.Exists)
                    {
                        string fullName = fileInfo.FullName.ToLower();
                        if (!ContainsEntryWithName(allFiles, fullName))
                        {
                            Entry e = new Entry(fullName, group.id);
                            allFiles.Add(e);
                        }
                    }
                }
            }
            if (group.directories.Count != 0)
            {
                if (group.extensions.Count == 0 && !group.extensions.Contains(".*"))
                    group.extensions.Add(".*");
            }
            else if (group.extensions.Count != 0)
            {
                if (group.directories.Count == 0 && !group.directories.Contains(workingDir))
                    group.directories.Add(workingDir);
            }
            foreach (string dir in group.directories)
            {
                string fullDir = dir;
                if (!IsPathAbsolute(fullDir))
                    fullDir = workingDir + "\\" + fullDir;
                DirectoryInfo dirInfo = new DirectoryInfo(fullDir);

                foreach (string ext in group.extensions)
                {
                    FileInfo[] files = dirInfo.GetFiles("*" + ext, SearchOption.AllDirectories);
                    foreach (FileInfo f in files)
                    {
                        // We're not considering files that were already processed.
                        if (destExtension.ToLower() == f.Extension.ToLower())
                            continue;

                        string fname = f.FullName.ToLower();
                        int groupId = group.id;
                        // If group is batched by directory, we divide this group
                        // into smaller groups, at most one group per unique directory.
                        if (group.batchByDirectory)
                        {
                            groupId = FindGroupIdContainingDir(f.DirectoryName);
                            if (groupId == -1)
                            {
                                groupId = entryGroups.Count;
                                EntryGroup fakeGroup = new EntryGroup();
                                fakeGroup.dir = f.DirectoryName;
                                fakeGroup.commandLine = group.commandLine;
                                fakeGroup.outDirectory = group.outDirectory;
                                fakeGroup.id = groupId;
                                entryGroups.Add(fakeGroup);
                            }
                        }

                        if (!ContainsEntryWithName(allFiles, fname))
                            allFiles.Add(new Entry(fname, groupId));
                    }
                }
            }
        }

        bool IsSourceNewer(string sourceFile, string destFile)
        {
            FileInfo srcFileInfo = new FileInfo(sourceFile);
            FileInfo dstFileInfo = new FileInfo(destFile);
            System.Diagnostics.Debug.Assert(srcFileInfo.Exists);
            // This should not happen, but some tools may output
            // temporary files that will be caught by *.* (default) extension.
            if (!srcFileInfo.Exists)
                return false;
            return !dstFileInfo.Exists || srcFileInfo.LastWriteTime > dstFileInfo.LastWriteTime;
        }
        bool IsPathAbsolute(string path)
        {
            // Very simple test, but it seems to be enough for our purposes.
            int driveSepIndex = path.IndexOf(':');
            return driveSepIndex >= 0 && driveSepIndex < 4;
        }
        string DestNameFromSourceName(string s, EntryGroup group)
        {
            // $$$ TODO: this can be easily done in a quicker, albeit
            // $$$ less portable way by searching for dot from the end
            // of the string. 
            FileInfo fi = new FileInfo(s);
            string ext = fi.Extension;

            string destName = s;
            if (group.outDirectory != null)
            {
                DirectoryInfo dirInfo = new DirectoryInfo(group.outDirectory);
                if (IsPathAbsolute(group.outDirectory))
                    destName = group.outDirectory + "\\" + fi.Name;
                else
                    destName = workingDir + "\\" + group.outDirectory + "\\" + fi.Name;
            }
            if (destName.IndexOf(ext) >= 0)
                return destName.Replace(ext, destExtension);
            // File doesnt contain extension.
            return destName + destExtension;
        }

        public Queue<Entry> GetFilesToProcess()
        {
            Queue<Entry> filesToProcess = new Queue<Entry>();
            foreach (Entry e in allFiles)
            {
                EntryGroup group = entryGroups[e.groupIndex];
                if (!lazyBuild || IsSourceNewer(e.fileName, DestNameFromSourceName(e.fileName, group)))
                    filesToProcess.Enqueue(e);
            }
            return filesToProcess;
        }

        public bool SendToolExeName(Socket s)
        {
            byte[] buffer = new byte[tool.Length + 10];
            buffer[0] = (byte)rde.Common.ServerToClientMsg.TOOL_EXECUTABLE;
            buffer[1] = ignoreErrors ? (byte)1 : (byte)0;
            System.Text.Encoding.ASCII.GetBytes(tool, 0, tool.Length, buffer, 2);
            return rde.Common.SendMessage(s, buffer, buffer.Length);
        }

        public class Entry
        {
            public Entry() { }
            public Entry(string fname, int group)
            {
                fileName = fname;
                groupIndex = group;
            }

            public void Send(Socket s)
            {
                byte[] buffer = new byte[fileName.Length + 10];
                buffer[0] = (byte)rde.Common.ServerToClientMsg.FILE_NAME;
                buffer[1] = (byte)groupIndex;
                System.Text.Encoding.ASCII.GetBytes(fileName, 0, fileName.Length,
                    buffer, 2);
                rde.Common.SendMessage(s, buffer, buffer.Length);
            }

            public string fileName;
            public int groupIndex;
        };
        public class EntryGroup
        {
            public void SendDescription(Socket s)
            {
                byte[] data = new byte[1 + 2 + 2 +
                    (outDirectory == null ? 0 : outDirectory.Length) + 
                    (commandLine == null ? 0 : commandLine.Length)];
                data[data.Length - 1] = 255;
                int i = 0;
                data[i++] = (byte)rde.Common.ServerToClientMsg.GROUP_DESCRIPTION;
                i += rde.Common.PutString(data, outDirectory, 1);
                i += rde.Common.PutString(data, commandLine, i);
                rde.Common.SendMessage(s, data);
            }   

            public List<String> extensions = null;
            public List<String> files = null;
            public List<String> directories = null;
            public string commandLine = null;
            public string outDirectory = null;
            public bool batchByDirectory = true;
            [XmlIgnore]
            public int id = 0;
            [XmlIgnore]
            public string dir = "";
        };
        // Groups should be sorted from "most specific" to "most general".
        // File names are most specific, then directories, then extensions.
        class EntryGroupComparer : IComparer<EntryGroup>
        {
            public int Compare(EntryGroup egx, EntryGroup egy)
            {
                if (egx.files.Count != 0 || egy.files.Count != 0)
                    return egy.files.Count - egx.files.Count;
                if (egx.directories.Count != 0 || egy.directories.Count != 0)
                    return egy.directories.Count - egx.directories.Count;
                if (egx.extensions.Count != 0 || egy.extensions.Count != 0)
                    return egy.extensions.Count - egx.extensions.Count;
                return 0;
            }
        }

        public void SendDescription(Socket s)
        {
            byte[] data = new byte[2 + 2 + destExtension.Length +
                2 + workingDir.Length + 2 + tool.Length + 2 + postBuildStep.Length +
                2 + postBuildArgs.Length];
            int i = 0;
            data[i++] = (byte)rde.Common.ServerToClientMsg.TASK_DESCRIPTION;
            data[i++] = ignoreErrors ? (byte)1 : (byte)0;
            i += rde.Common.PutString(data, destExtension, i);
            i += rde.Common.PutString(data, workingDir, i);
            i += rde.Common.PutString(data, tool, i);
            i += rde.Common.PutString(data, postBuildStep, i);
            i += rde.Common.PutString(data, postBuildArgs, i);
            rde.Common.SendMessage(s, data);
        }

        void SendBuildStep(Socket s, string buildStep, string buildStepArgs)
        {
            if (buildStep != null && buildStep.Length > 0)
            {
                byte[] data = new byte[1 + 2 + buildStep.Length + 2 + buildStepArgs.Length];
                int i = 0;
                data[i++] = (byte)rde.Common.ServerToClientMsg.START_PROCESS;
                i += rde.Common.PutString(data, buildStep, i);
                i += rde.Common.PutString(data, buildStepArgs, i);
                rde.Common.SendMessage(s, data);
            }
        }
        public void SendPreBuildStep(Socket s)
        {
            SendBuildStep(s, preBuildStep, preBuildArgs);
        }

        public void SendGroupDescriptions(Socket s)
        {
            foreach (EntryGroup group in entryGroups)
                group.SendDescription(s);
        }

        public void PostProcess()
        {
            entryGroups.Sort(new EntryGroupComparer());
            for (int i = 0; i < entryGroups.Count; ++i)
                entryGroups[i].id = i;
        }

        public string name = "";
        public string description = "";
        public string preBuildStep = "";
        public string preBuildArgs = "";
        public string postBuildStep = "";
        public string postBuildArgs = "";
        public string workingDir = ".";
        public List<EntryGroup> entryGroups = new List<EntryGroup>();
        public List<Entry> allFiles = new List<Entry>();
        public string destExtension = "";
        public string tool = "";
        public bool lazyBuild = true;
        public bool batchProcessing = false;
        public bool ignoreErrors = false;
    }
}
